import logging
import re

log = logging.getLogger('router')

CIDR_REGEX = re.compile("^!?([0-9A-Fa-f]{2,8})$"
                        "|^!?([0-9A-Fa-f]{2,8})/$"
                        "|^!?([0-9A-Fa-f]{2,8})/[0-9]{1,2}$")


class Interval():
    bitlen = 32

    def __init__(self, arg1=None, arg2=None):
        if arg2 is not None:
            self.minimum = int(arg1)
            self.maximum = int(arg2)
        else:
            if arg1 is not None:
                _arg = arg1
                if "-" == arg1[0]:
                    self.reverse = True
                    _arg = arg1[1:]
                if "+" == arg1[0]:
                    self.reverse = False
                    _arg = arg1[1:]
                if "/" in _arg:
                    B, M = _arg.split("/")
                    if M == "":
                        M = str(Interval.bitlen)
                else:
                    B = _arg
                    M = str(Interval.bitlen)
                baseAddr = int(B, 16)
                maskLen = int(M)
                mask = (pow(2, Interval.bitlen) - 1) >> maskLen
                self.minimum = baseAddr & ~mask
                self.maximum = baseAddr | mask
            else:
                self.minimum = 0
                self.maximum = (pow(2, Interval.bitlen) - 1)  # 0xffffffff

    def show(self):
        log.debug("["+hex(self.minimum)+","+hex(self.maximum)+"]")

    def contains(self, _value):
        return (self.minimum <= _value) and (self.maximum >= _value)

    def get_interval(self):
        return [self.minimum, self.maximum]

    def get_intersection(self, _interval):
        intersection = Interval(max(self.minimum, _interval.minimum),
                                min(self.maximum, _interval.maximum))

        if intersection .maximum < intersection.minimum:
            return None
        return intersection

    def get_union(self, _interval):
        min_minimum = min(self.minimum, _interval.minimum)
        min_maximum = min(self.maximum, _interval.maximum)
        max_minimum = max(self.minimum, _interval.minimum)
        max_maximum = max(self.maximum, _interval.maximum)
        intersection = self.get_intersection(_interval)

        if (intersection is not None) or (min_maximum + 1 == max_minimum):
            return [Interval(min_minimum, max_maximum)]

        return [Interval(min_minimum, min_maximum),
                Interval(max_minimum, max_maximum)]

    def get_difference(self, _interval):
        intersection = self.get_intersection(_interval)
        if intersection is None:
            return [Interval(self.minimum, self.maximum)]

        if (_interval.minimum <= self.minimum and
                _interval.maximum >= self.maximum):
            return []

        if (_interval.minimum > self.minimum and
                _interval.maximum < self.maximum):
            return [Interval(self.minimum, _interval.minimum - 1),
                    Interval(_interval.maximum + 1, self.maximum)]

        if (_interval.minimum <= self.minimum):
            return [Interval(_interval.maximum + 1, self.maximum)]
        else:
            return [Interval(self.minimum, _interval.minimum - 1)]


class CidrInterval(Interval):
    def __init__(self, arg1=None, arg2=None):
        _arg1 = arg1
        self.reverse = False
        if (type(arg1) is str) and (arg1 is not None) and (arg1[0] == '!'):
            _arg1 = arg1[1:]
            self.reverse = True
        super().__init__(_arg1, arg2)

    def show(self):
        log.debug("Reverse = " + str(self.reverse))
        super().show()


class Set():
    def __init__(self):
        self.clear()

    def clear(self):
        self.m_intervals = []  # [CidrInterval()]
        self.m_virgin = True

    def show(self):
        log.debug(" --- SET ---")
        for interval in self.m_intervals:
            interval.show()
        log.debug("")

    def get_intervals(self):
        retVal = []
        for interval in self.m_intervals:
            retVal.append(interval.get_interval())
        return retVal

    def contains(self, _value):
        for interval in self.m_intervals:
            if (interval.contains(_value)):
                return True
        return False

    def do_union(self, _interval):
        remaining_interval = _interval
        new_intervals = []

        for interval in self.m_intervals:
            union_intervals = interval.get_union(remaining_interval)

            if (len(union_intervals) == 2):  # Disjoint intervals
                # new_intervals.push_back(union_intervals.front())
                new_intervals.append(union_intervals[0])

            remaining_interval = union_intervals[-1]

        new_intervals.append(remaining_interval)
        self.m_intervals = new_intervals
        self.m_virgin = False

    def do_difference(self, _interval):
        new_intervals = []

        if (self.m_virgin is True):
            self.m_intervals.append(CidrInterval())

        for interval in self.m_intervals:
            for new_interval in interval.get_difference(_interval):
                new_intervals.append(new_interval)

        self.m_intervals = new_intervals
        self.m_virgin = False

    def load_cidr_intervals(self, _cidrs):
        for filter in _cidrs:
            if CIDR_REGEX.match(filter) is None:
                log.warning("Invalid filter : " + filter)
                continue
            log.debug("inserting filter " + filter)
            try:
                interval = CidrInterval(filter)
                interval.show()
                if (interval.reverse):
                    self.do_difference(interval)
                else:
                    self.do_union(interval)
            except Exception as e:
                log.error("Invalid filter: \"" + str(filter) + "\": " + str(e))

    def load_interval(self, first, last):
        log.debug("inserting interval [" + hex(first) + "," + hex(last) + "]")
        try:
            interval = CidrInterval(first, last)
            interval.show()
            if (interval.reverse):
                self.do_difference(interval)
            else:
                self.do_union(interval)
        # except Exception as e:
        except MemoryError as e:
            log.error("Invalid filter: \"" + str(filter) + "\": " + str(e))
