import argparse
import configparser
import filecmp
import json
import logging.config
import os
import socket
import ssl
from shutil import copyfile

from keroslib import utils

# Arguments priority list
# 1- Command line arguments
# 2- Environment variable
# 3- `station.conf`
# 4- default
# Default files, following folder priority:
# 1- STATION_HOME_DIR (default: local folder ".")
# 2- ${HOME}/.config/station/
# 3- /etc/station/

log = logging.getLogger('config')

default = {
    'homedir': ".",
    'tempdir': "/tmp",
    'lorad': {
        'uplink': "ipc:///var/run/lora/uplink",
        'downlink': "ipc:///var/run/lora/downlink",
        'control': "ipc:///var/run/lora/lorad",
    },
    'router': {},
    'cups': {},
    'station': {},
    'database': {},
    'log': {
        'log_level': "DEBUG",
        'log_file': "stdout",
    }
}

# declare a global variable to help the backup
# function find the credentials files
confdir = None


def readconfig(cred_suffix=""):
    global confdir
    # Version from local version.txt
    versionfile = os.path.dirname(os.path.realpath(__file__)) + "/version.txt"
    try:
        with open(versionfile) as fd:
            version = fd.readline().strip()
    except FileNotFoundError:
        version = "unknown"
    log.debug("gathering '" + cred_suffix + "' credentials")
    parser = argparse.ArgumentParser(
        description="Kerlink's station",
        add_help=False,  # -h is for --home
        argument_default=argparse.SUPPRESS
    )
    # Help arguments
    parser.add_argument("-?", "--help", help="Give this help list",
                        action="help")
    parser.add_argument("--usage", help="Give a short usage message",
                        action="store_true")
    parser.add_argument("-v", "--version", help="Print station version",
                        action="version", version="%(prog)s " + version)
    parser.add_argument("-p", "--params", help="Print current parameter"
                        " settings",
                        action="store_true")

    # Directory arguments
    parser.add_argument("-h", "--home",
                        help="Home directory for configuration files. "
                        "Default is /etc/station.",
                        metavar='DIR')

    # Log arguments
    parser.add_argument("-L", "--log_file",
                        help="Write log entries to FILE. If FILE is '-' then "
                        "write to stderr. If FILE is syslog, write logs using "
                        " syslog.",
                        metavar='FILE')

    args = parser.parse_args()
    args = vars(args)

    if "usage" in args:
        parser.print_usage()
        parser.exit(0)

    # homedir is necessary to get `station.conf` values and URIs files
    homedir = args.get('home', os.getenv("STATION_DIR", "."))

    # Default folders, in priority order (first prefered) for ini and uri files
    default_folders = [homedir,
                       os.path.expanduser("~/.config/station"),
                       "/etc/station"]

    # Retrieve configuration from file
    ini = configparser.ConfigParser()
    # Configuration file search (in configparser, latest path is preferred)
    inipaths = default_folders[::-1]
    inifiles = ini.read([path + "/station.conf" for path in inipaths])

    config = {}
    config['version'] = version
    config['homedir'] = homedir
    confdir = homedir
    config['tempdir'] = args.get('temp', os.getenv(
        "STATION_TEMP", ini.get("DEFAULT", "tempdir", fallback="/tmp")))
    # lorad config
    config['lorad'] = {}
    ignreconf = ini.getboolean("lorad", "ignore_reconf", fallback=False)
    config['lorad']['ignore_reconf'] = ignreconf

    for uri in ['uplink', 'downlink', 'control']:
        config['lorad'][uri] = ini.get(
            "lorad", uri, fallback=default['lorad'][uri])

    # router and cups config
    config['router'] = {}
    config['cups'] = {}

    # files: {tc,cups}.{uri,trust,crt,key}
    # router = tc; cups (original code and file names kept for consistency)
    # *.uri  : server's address
    # *.trust: server's CA certificate
    # *.crt  : private Station certificate
    # *.key  : private Station key or Token

    services = [('router', 'tc'+cred_suffix), ('cups', 'cups'+cred_suffix)]
    fileext = ['trust', 'crt', 'key']

    for cfg, service in services:
        # 1- search .uri file in default_folders, in order
        fname = '.'.join((service, 'uri'))
        uri = None
        for folder in default_folders:
            filename = os.path.join(folder, fname)
            if os.path.isfile(filename) and os.path.getsize(filename) != 0:
                with open(filename) as fd:
                    uri = fd.read().strip()
                    break
        config[cfg]['uri'] = uri
        if not uri:
            # URI file not found: set credentials to `None` and move-on
            for ext in fileext:
                config[cfg][ext] = None
            continue

        # service_folder = folder
    # 2- search other files in that folder
        for ext in fileext:
            fname = '.'.join((service, ext))
            config[cfg][ext] = None
            filename = os.path.join(folder, fname)
            if os.path.isfile(filename) and os.path.getsize(filename) != 0:
                config[cfg][ext] = filename

    # Get board infos
    if os.path.isfile("/tmp/board_info.json"):
        with open("/tmp/board_info.json") as fd:
            board_info = json.load(fd)
    else:
        board_info = {}

    # Router ID
    routerid = ini.get("router", "id", fallback=None)
    if not routerid:
        routerid = int(utils.getEUI64(), 16)

    config['router']['id'] = routerid

    # Websockets Ping Interval
    ping_interval = ini.get("router", "ping_interval", fallback="60")
    if ping_interval.upper() == "NONE":
        ping_interval = None
    else:
        # convert to int here rather than on 'ini.get' to handle 'None' option
        ping_interval = int(ping_interval)
    config['router']['ping_interval'] = ping_interval

    # TCP Socket Timeout (5s minimum)
    sock_tmo = max(int(ini.get("router", "sock_tmo", fallback=20)), 5)
    config['router']['sock_tmo'] = sock_tmo

    # Default Net ID to filter on
    filters = ini.get("router", "filters", fallback="NONE")
    if filters.upper() == "NONE":
        filters = None
    config['router']['filters'] = filters

    # Land Mobile Station Standard
    jplms = ini.getboolean("router", "use_JPLMS_standard", fallback=False)
    config['router']['jplms'] = jplms

    # CUPS interval
    config['cups']['interval'] = int(
        ini.get("cups", "interval", fallback=3600))

    # CUPS capath
    config['cups']['capath'] = ini.get(
        "cups", "capath", fallback=ssl.get_default_verify_paths().capath)

    # station config
    config['station'] = {}
    # Station model (board radio type)
    if 'cpu' in board_info and 'radio' in board_info['cpu']:
        model = board_info['cpu']['radio'].get('LORABOARD_TYPE')
    else:
        if 'lpbs' in socket.gethostname():
            reg = "???"
            for m in range(len(board_info['module'])):
                bim = board_info['module'][m]
                if 'frontend' in bim:
                    # use region of the first FE module found
                    reg = bim['frontend']['board_freq'][:3]
                    break
            model = 'iBTS-'+reg
        elif 'zeth' in socket.gethostname():
            reg = "???"
            if "DP" in board_info['cpu']['board_label']:
                reg = "868"
            if "DQ" in board_info['cpu']['board_label']:
                reg = "915"
            model = 'iZepto_Eth-'+reg
        elif 'zcel' in socket.gethostname():
            reg = "???"
            if "DW" in board_info['cpu']['board_label']:
                reg = "868"
            if "DX" in board_info['cpu']['board_label']:
                reg = "915"
            model = 'iZepto_Cell-'+reg
        else:
            model = "unknown"
    config['station']['model'] = model

    # database config
    config['database'] = {}
    dben = ini.getboolean("database", "enable", fallback=False)
    config['database']['enable'] = dben
    depth = int(ini.get("database", "depth", fallback=20000))
    config['database']['depth'] = depth
    dbpath = ini.get("database", "dbpath",
                     fallback="/var/lib/basic_station/station_database.sqlite")
    config['database']['dbpath'] = dbpath

    showdelayed = ini.getboolean("database", "show_delayed", fallback=False)
    config['router']['show_delayed'] = showdelayed

    # Firmware version
    if os.path.isfile("/etc/version"):
        with open("/etc/version") as fd:
            config['station']['firmware'] = fd.read().strip()
    else:
        config['station']['firmware'] = "unknown"

    # log config
    config['log'] = {}
    # Log levels
    log_level = {}
    # Command line and ENV parsing
    level = args.get('log_level', os.getenv("STATION_LOGLEVEL"))
    if level:
        log_level = parse_log_level(level)

    # 'station.conf' parsing
    for mod in ['lorad', 'router', 'cups', 'station', 'database', 'DEFAULT']:
        if mod in log_level:
            # Command line or env already defines log level
            continue
        log_level[mod] = ini.get(
            mod, "log_level", fallback=log_level.get('DEFAULT', "INFO"))

    config['log']['level'] = log_level

    # Log file
    log_file = args.get('log_file', os.getenv("STATION_LOGFILE"))
    log_size = log_rot = None
    if log_file:
        logconf = log_file.split(",")
        if len(logconf) == 1:
            log_file = logconf[0]
        elif len(logconf) == 2:
            log_file, log_size = logconf
        else:
            log_file, log_size, log_rot = logconf
    if log_file is None:
        log_file = ini.get("log", "file", fallback="stdout")
    if log_size is None:
        log_size = ini.get("log", "size", fallback=0)
    if log_rot is None:
        log_rot = ini.get("log", "rotate", fallback=0)

    if log_file == "-":
        log_file = "stderr"
    config['log']['file'] = log_file
    config['log']['size'] = int(log_size)
    config['log']['rot'] = int(log_rot)

    # Setup logger
    # We support a single handler, it should handle all levels, hence "DEBUG"
    log_handler = {
        "level": "DEBUG",
        "formatter": "standard",
    }
    # Default formatter
    # formatter = "%(asctime)s [%(name).8s:%(levelname)s] %(message)s"

    if config['log']['file'] in ["stdout", "stderr"]:
        log_handler.update({"class": "logging.StreamHandler",
                            "stream": "ext://sys." + config['log']['file']})
    elif config['log']['file'] in ["syslog", "rsyslog"]:
        log_handler.update({"class": "logging.handlers.SysLogHandler",
                            "address": "/dev/log",
                            "formatter": "syslog"})
    else:
        log_handler.update({"class": "logging.handlers.RotatingFileHandler",
                            "filename": config['log']['file'],
                            "maxBytes": config['log']['size'],
                            "backupCount": config['log']['rot']})

    log_logger = {
        "handlers": ["default"],
        "propagate": False
    }

    # Configure logger
    logconfig = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {
                "format": "%(asctime)s [%(name).8s:%(levelname)s] %(message)s"
            },
            "syslog": {
                "format": "station [%(name).8s:%(levelname)s] %(message)s"
            },
        },
        "handlers": {"default": log_handler},
        "loggers": {}
    }
    logconfig["loggers"][""] = {**log_logger,
                                "level": log_level.get('DEFAULT', "INFO")}
    for mod in log_level:
        logconfig["loggers"][mod] = {**log_logger, "level": log_level[mod]}

    logging.config.dictConfig(logconfig)

    # Print current parameter settings and exit
    if "params" in args:
        import pprint
        pprint.pprint(config)
        parser.exit(0)

    # Delayed ini file logging after loggers have been configured
    log.info(f"Starting BasicStation daemon {version}")
    if not inifiles:
        log.warning(
            f"No configuration file found in {inipaths}. Fallback to default")
    else:
        log.info(f"Read configuration from {inifiles}")
    log.debug(config)

    return config


def parse_log_level(log_level):
    '''
    Parse a log level command such as: "DEBUG,SYS:INFO,RAL:XDEBUG"
    format: module:level,... where module is optional
    '''
    log_norm_levels = ["XDEB", "DEBU", "VERB",
                       "INFO", "NOTI", "WARN", "ERRO", "CRIT"]
    log_avai_levels = ["DEBUG", "DEBUG", "DEBUG", "INFO", "WARNING", "WARNING",
                       "ERROR", "CRITICAL"]
    log_modules = {
        "any": "DEFAULT",
        "ral": "lorad",
        "s2e": "station",
        "wss": "websockets",
        "aio": "asyncio",
        "cup": "cups",
        "sys": "system",
        "tce": "router",
    }

    logs = {}
    for level in log_level.split(','):
        try:
            m, level = level.split(':')
        except ValueError:
            m = "any"  # level without module

        m = m.lower()
        if m in log_modules:
            m = log_modules[m]
        if m not in log_modules.values():
            # Unknown module
            continue
        if not level.isdigit():
            level = log_norm_levels.index(level.upper()[:4])
        logs[m] = log_avai_levels[int(level)]

    return logs


def backup(service, suffix='-bak', revert=False):
    global confdir

    for ext in ('trust', 'crt', 'key', 'uri'):
        filename = os.path.join(confdir, ".".join((service, ext)))
        # ["/path/to/cups", ".crt"]
        dest = list(os.path.splitext(filename))
        # ["/path/to/cups", "-bak", ".crt"]
        dest.insert(1, suffix)
        # "/path/to/cups-bak.crt"
        dest = "".join(dest)

        if os.path.isfile(filename) and os.path.getsize(filename) != 0:
            # back-up only if files differ
            if (not os.path.isfile(dest)) or \
               (filecmp.cmp(filename, dest) is False):
                if revert is False:
                    log.info("backuping " + str(filename))
                    copyfile(filename, dest)
                else:  # reverting onto invalid file
                    log.info("reverting onto " + str(dest))
                    copyfile(dest, filename)
        else:
            if revert is True and os.path.isfile(dest):
                # reverting onto missing file
                log.info("creating " + str(filename) + " from " + str(dest))
                copyfile(dest, filename)
