import asyncio
import base64
import hashlib
import logging
import os
import signal
import sys

from . import config
from . import cups
from . import database
from . import lorad
from . import router
from . import stats

if sys.version_info[1] < 7:
    asyncio.create_task = asyncio.ensure_future
    asyncio.all_tasks = asyncio.Task.all_tasks
    asyncio.current_task = asyncio.Task.current_task

log = logging.getLogger('station')

db_size = 0


async def unbufferize_task(rt, db_client):
    global db_size
    log.debug("Starting Unbufferize Task")
    await rt.connected.wait()
    # do not send buffered messages until
    # router_config message is received
    await rt.router_config.wait()

    # make sure router_config params have all been processed
    while not hasattr(rt, 'DRs'):
        log.debug("router DRs not configured ..")
        await asyncio.sleep(1)

    # give some time to LNS to be ready to handle several TX in a row
    # (necessary with at least Actility server)
    await asyncio.sleep(2)

    # loop in case the DB depth parameter has been decreased in the config
    # file and there are more messages than this new value
    db_size = await db_client.size()

    while db_size > 0:
        # retrieve only 5 messages at a time to allow
        # other tasks to be scheduled while unbufferizing
        MSGS = await db_client.fetch(5)
        log.info(str(len(MSGS)) + " buffered messages to send")

        for [ID, M] in MSGS:
            await rt.uplink(*M, delayed=True)
            if await db_client.erase(ID) != 0:
                log.warning("message erasing failure, buffer not empty")
                return
        await asyncio.sleep(0.02)

        db_size = await db_client.size()

    log.debug("Unbufferize Task Complete")


async def uplink_task(rt, ld, versions, dbcfg):
    global db_size
    dbenabled = dbcfg['enable']
    dbpath = dbcfg['dbpath']
    dbmaxsize = dbcfg['depth']
    unbuff_task = None

    '''Handle uplink messages: get them from lorad and send to router'''
    try:
        await asyncio.wait_for(rt.send_version(versions), timeout=10)
    except asyncio.TimeoutError:
        log.warning("router send version timeout")

    recv_tasks = [asyncio.create_task(
        ld.recv('lora')), asyncio.create_task(ld.recv('gfsk'))]

    if dbenabled is True:
        db_client = database.DatabaseClient(dbmaxsize, dbpath)
        await db_client.create_connection()
        unbuff_task = asyncio.create_task(unbufferize_task(rt, db_client))
    else:
        log.info("Database is disabled")

    if os.environ.get('INJECTOR_PORT') is not None:
        asyncio.create_task(rt.injector_task(
            int(os.environ.get('INJECTOR_PORT'))))

    while True:
        done_tasks, pending_tasks = \
            await asyncio.wait(recv_tasks, return_when=asyncio.FIRST_COMPLETED)
        # Backup pending tasks in current receive task list
        recv_tasks = list(pending_tasks)
        for done_task in done_tasks:
            msg = done_task.result()
            if not msg:
                log.warning("Unexpected behaviour, no message received")
                continue

            if 'lora' in msg:
                msg_type = 'lora'
            elif 'gfsk' in msg:
                msg_type = 'gfsk'
            else:
                log.warning("Unhandled message type")
                continue

            # Append new task since previous is finished to receive task list
            recv_tasks.append(asyncio.create_task(ld.recv(msg_type)))

            if 'data' not in msg[msg_type]:
                log.warning(f"No data in {msg_type} message")
                continue

            data = msg[msg_type]['data']
            if data['crc'] != "CRC_OK":
                log.warning(f"CRC not OK {msg_type} frame received: discard")
                continue

            radio = msg[msg_type]['radios'][0]  # FIXME: browse radio list
            rec_time = None

            if msg_type == 'lora':
                # Translate SF/BW from string to int :
                # "SF_7" -> 7; "BW_125K" -> 125
                sf = int(
                    ''.join(c for c in data['spreadingFactor'] if c.isdigit()))
                bw = int(''.join(c for c in data['bandwidth'] if c.isdigit()))
                if "time" in radio['demodulator']:
                    # see http://www.leapsecond.com/java/gpsclock.htm
                    # for values explanations
                    Delta = 0
                    time_format = radio['demodulator']['time']['format']
                    if time_format == 'TAI_TIME':
                        Delta = +19  # TAI to GPS
                    elif time_format == 'UNIX_TIME':
                        Delta = -18  # UTC to GPS
                    microsec = 0
                    if "microseconds" in radio['demodulator']['time']:
                        microsec = radio['demodulator']['time']['microseconds']
                    rec_time = ((radio['demodulator']['time']['time']
                                - 315964800 - Delta) * 1000000 +
                                microsec)
                snr = radio['demodulator']['snr']['snr']
            elif msg_type == 'gfsk':
                # Corresponding SF/BW in DRs table are 0/0
                sf = 0
                bw = 0
                snr = None

            payload = base64.b64decode(data['payload'])

            if (not rt.connected.is_set()) or (db_size > 0):
                # log.warning(" !!!! WE SHOULD BUFFERIZE THIS !!!!")
                if dbenabled is True:
                    await db_client.insert(payload, data['frequency'], sf, bw,
                                           radio['radioId'],
                                           radio['demodulator']['counter'],
                                           radio['demodulator']['rssi'], snr,
                                           rcvtime=rec_time)
                    if unbuff_task is None:
                        unbuff_task = asyncio.create_task(
                            unbufferize_task(rt, db_client))

            else:
                if (unbuff_task is not None):
                    try:
                        db_size = await db_client.size()
                        # give the task up to 15 minutes to complete
                        await asyncio.wait_for(unbuff_task, timeout=900)
                        log.debug("unbuff_task completed")
                    except TimeoutError:
                        log.warning(
                            "Unbufferize task should have completed by now !!")
                    log.debug("clearing unbuff_task")
                    unbuff_task = None

                await rt.uplink(payload, data['frequency'], sf, bw,
                                radio['radioId'],
                                radio['demodulator']['counter'],
                                radio['demodulator']['rssi'], snr,
                                rcvtime=rec_time)


async def downlink_task(rt: router.RouterClient, ld: lorad.Lorad, ldcfg: dict):
    prev_router_config_hash = None
    '''Handle downlink messages: get them from router and send to lorad'''
    while True:
        msg = await rt.recv()
        if not msg:
            continue
        log.debug(f"New downlink {msg}")

        if msg['msgtype'] == 'dnmsg':
            asyncio.create_task(send_and_ack(rt, ld, msg))
        elif msg['msgtype'] == 'router_config':
            # compute a MD5 hash on the significant lorad settings
            # to work out if a new lorad restart is really required
            conf_concat = ""
            sections = ["hwspec", "sx1301_conf",
                        "freq_range", "bcning", "region"]
            for section in sections:
                try:
                    conf_concat = conf_concat + str(msg[section]) + "##"
                except Exception:
                    conf_concat = conf_concat + "_NO_" + section + "##"
            md5 = hashlib.md5(conf_concat.encode('utf-8')).hexdigest()
            log.debug("config MD5 : " + str(md5))
            if prev_router_config_hash is not None:
                if prev_router_config_hash == md5:
                    log.info("no config diff detected, skipping lorad update")
                    continue
            prev_router_config_hash = md5

            # hwspec = sx1301/N ; Test N
            nb_chips = int(msg['hwspec'].split('/')[1])
            if nb_chips > 1:
                log.warning("Unable to manage a number of chips > 1")
                continue

            if ldcfg['ignore_reconf']:
                log.info("Ignore lorad reconfiguration")
            else:
                for n in range(0, nb_chips):
                    await ld.apply_radio_conf(rt.region, msg['sx1301_conf'][n])
                if 'freq_range' in msg:
                    await ld.apply_freq_range(msg['freq_range'])
                if 'bcning' in msg and msg['bcning'] is not None:
                    if 'region' in msg:
                        await ld.apply_beaconing_conf(msg['bcning'],
                                                      msg['region'])
                    else:
                        log.warning("Cannot apply beaconing without\
                                      region info")
                else:
                    await ld.apply_beaconing_conf(None, rt.region)
                # apply all configuration changes
                await ld.apply_changes()

        elif msg['msgtype'] == 'dnsched':
            asyncio.create_task(send_schedule(rt, ld, msg))
        else:
            log.warning("Unhandled message type \"{msgtype}\"".format(**msg))


def build_tx_config(rt: router.RouterClient, msg: dict, counter: int, rx_id):
    # Transmission configuration

    # Get max TX power allowed by regulation
    power = lorad.get_max_tx_power(rt.region, msg[rx_id + "Freq"], rt.jplms)

    # Override if max_eirp from router_config is lower
    if rt.max_eirp is not None and rt.max_eirp < power:
        power = rt.max_eirp

    # LNS protocol specifies lowest priority: 0          and highest: 255
    # Lorad interprets       lowest priority: 0xFFFFFFFF and highest: 1
    # Map highest LNS to highest lorad
    priority = 256 - msg["priority"]
    sf, bw, dwonly = rt.DRs[msg[rx_id + "DR"]]

    rx = {'rid': msg["rctx"], 'frq': msg[rx_id + "Freq"], 'sf': sf,
          'bw': bw, 'pwr': power, 'cnt': counter, 'prio': priority}

    if 1 == msg['dC']:  # class B
        rx['gpstime'] = msg['gpstime']

    log.debug(f"Tx Config : {rx}")
    return rx


async def send_and_ack(rt, ld, msg):
    '''Actually send a downlink message to lorad and wait for
       the transmission event. When transmission is successful,
       acknowledge to router '''

    rt.stats_dn_msg_recv = rt.stats_dn_msg_recv + 1

    args = {}
    # limit msgid to 32 bits for lorad API compatibility
    args['msgid'] = msg.get("diid", 0) & 0xFFFFFFFF

    args['payload'] = bytes.fromhex(msg.get("pdu", b''))
    if 'RxDelay' in msg:
        if ('xtime' in msg) and (msg['RxDelay'] > 0):
            counter = msg['xtime'] + 1000000 * msg['RxDelay']
        else:
            # this could happen for Class C downlinks
            del msg['RxDelay']
            counter = None
    else:
        counter = None

    if 'rctx' not in msg:
        # this is an optional field but we need it
        msg['rctx'] = 0

    if 1 == msg['dC']:
        # Class B
        counter = None
        msg['RX2DR'] = msg['DR']
        msg['RX2Freq'] = msg['Freq']
        log.debug(f"Patching Class B Message: {msg}")

    # Browse RX1/RX2 parameters
    for rx_id in ("RX1", "RX2"):
        if rx_id + "DR" in msg and rx_id + "Freq" in msg:
            if (rx_id == "RX2") and ('RxDelay' in msg):
                counter += 1000000  # RX1 + 1s
            args['txconfig'] = build_tx_config(rt, msg, counter, rx_id)

            log.info("Sending downlink {msgid} on RX ".format(**args) + rx_id)

            dl, event = await asyncio.gather(ld.send(**args),
                                             ld.wait_txevent(args["msgid"]))
            # If transmission succeeds on RX1, don't try on RX2
            if event["transmission"]["status"] == "TRANSMITTED":
                log.info("Successful transmission {diid}".format(**msg))
                # None = immediate
                dl_counter = None
                if 'RxDelay' in msg:
                    dl_counter = dl["lora"]["configurations"][0]["timings"][0]["counter"]  # noqa: E501
                await rt.dntxed(msg["diid"], msg["DevEui"],
                                msg["rctx"], dl_counter,
                                event["transmission"]["txtime"])
                if rx_id == "RX1":
                    rt.stats_dn_rx1 = rt.stats_dn_rx1 + 1
                if rx_id == "RX2":
                    rt.stats_dn_rx2 = rt.stats_dn_rx2 + 1

                break
            else:
                if (rx_id == "RX1"):
                    log.warning("Transmission {} status: {}".format(
                                msg["diid"], event["transmission"]["status"]))
                else:
                    log.error("Transmission {} status: {}".format(
                              msg["diid"], event["transmission"]["status"]))


async def send_schedule(rt, ld, mmsg):
    '''Send a downlink message to lorad and wait for
       the transmission event. '''

    # Sort the list of downlinks by ascending transmission times
    sorted_list = sorted(mmsg['schedule'], key=lambda d: d['gpstime'])

    for _msg_id_, msg in enumerate(sorted_list, start=1):

        # mimic class B setting to allow TX at a given timestamp
        msg['dC'] = 1
        counter = None

        args = {}
        args['payload'] = bytes.fromhex(msg.get("pdu", b''))
        # limit msgid to 32 bits for lorad API compatibility
        args["msgid"] = _msg_id_ & 0xFFFFFFFF

        if 'rctx' not in msg:
            # this is an optional field but we need it
            msg['rctx'] = 0

        # manually configure an RX window
        rx_id = "RX2"
        msg['RX2DR'] = msg['DR']
        msg['RX2Freq'] = msg['Freq']

        args['txconfig'] = build_tx_config(rt, msg, counter, rx_id)

        log.info("Sending downlink schedule #" + str(_msg_id_))

        dl, event = await asyncio.gather(ld.send(**args),
                                         ld.wait_txevent(args["msgid"]))

        if event["transmission"]["status"] == "TRANSMITTED":
            log.info(
                "Successful scheduled transmission #{msgid}".format(**args))
        else:
            log.error("Schedule transmission status: {}".format(
                event["transmission"]["status"]))


async def cups_task(cu, rt, cuuri, cutrust, cucrt,
                    cukey, capath, interval, versions):
    '''CUPS main loop'''
    tcuri, tctrust, tccrt, tckey = rt.get_credentials()
    # rollback levels in case of lack or credentials or corruption
    rb_level = 0
    rb = ["", "-bak", "-boot"]
    autobackup = True
    while True:
        if not cuuri:  # this can happen if "-bak" credentials do not exist
            updated = None
        else:
            updated = await cu.update_info(cuuri, cutrust, cucrt, cukey,
                                           tcuri, tctrust, tccrt, tckey,
                                           capath, versions,
                                           auto_backup=autobackup)

        if updated is None:
            rb_level = rb_level + 1

            if rb_level >= len(rb):
                log.info("No more credentials available to establish " +
                         "connection to a CUPS server. Retrying later")
                await asyncio.sleep(interval)
                rb_level = 0
                autobackup = True
            else:
                log.warning("Something abnormal happened. Switching to " +
                            rb[rb_level] + " credentials set.")
                # disable automatic backup as we might be able to rollback
                autobackup = False

            rollback_cfg = config.readconfig(rb[rb_level])['cups']
            cuuri = rollback_cfg['uri']
            cutrust = rollback_cfg['trust']
            cucrt = rollback_cfg['crt']
            cukey = rollback_cfg['key']

            await asyncio.sleep(2)
            continue

        if updated['cups']:
            cuuri = cu.uri['cups']
            cutrust, cucrt, cukey = map(
                cu.cred['cups'].get, ('trust', 'crt', 'key'))
        if updated['tc']:
            # Router information update: re-connect with the new set.
            # clean router tcuri to prevent reconnection to previous URI
            rt.tcuri = None
            await rt.disconnect()
            tcuri = cu.uri['tc']
            tctrust, tccrt, tckey = map(
                cu.cred['tc'].get, ('trust', 'crt', 'key'))
            await rt.connect(tcuri, tctrust, tccrt, tckey, capath)
            # issue a send version to be certain to gather a new
            # router-config message
            if rt.connected.is_set():
                await rt.send_version(rt.versions)

        # reset roll-back level for next time and make sure that
        # first level credentials are up to date
        if rb_level > 0:
            await cu.roll_back(rb[rb_level])

        rb_level = 0
        autobackup = True

        if not updated['cups'] and not updated['tc']:
            await asyncio.sleep(interval)
        else:
            # do not send a new update_info right away
            await asyncio.sleep(10)


async def get_versions(cfg, ld):
    '''Format version strings for CUPS and router'''
    station = cfg['version'] + " (keros/lorad)"
    firmware = cfg['station']['firmware']
    package = await ld.versions()
    model = cfg['station']['model']
    return {"station": station, "firmware": firmware, "package": package,
            "model": model}


# Runtime
async def shutdown(loop, signal=None):
    """Cleanup tasks tied to the service's shutdown."""
    if signal:
        log.warning(f"Received exit signal {signal.name}...")
    tasks = [t for t in asyncio.all_tasks() if t is not
             asyncio.current_task()]

    [task.cancel() for task in tasks]

    log.debug(f"Cancelling {len(tasks)} outstanding tasks")
    await asyncio.gather(*tasks, return_exceptions=True)
    loop.stop()


def handle_exception(loop, context):
    # Run default exception handler (log Traceback as asyncio:ERROR)
    loop.default_exception_handler(context)
    # context["message"] will always be there; but context["exception"] may not
    msg = context.get("exception", context["message"])
    log.error(f"Caught exception: {msg}")
    log.info("Shutting down...")
    asyncio.create_task(shutdown(loop))


def log_except_hook(exc_type, exc_value, exc_traceback):
    # Do not print exception when user cancels the program
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    log.error("An uncaught exception occurred:")
    log.error("Type: %s", exc_type)
    log.error("Value: %s", exc_value)

    if exc_traceback:
        log.debug("", exc_info=(exc_type, exc_value, exc_traceback))


def main():
    sys.excepthook = log_except_hook

    cfg = config.readconfig()

    rtcfg = cfg['router']
    rt = router.RouterClient(rtcfg)
    ldcfg = cfg["lorad"]
    ld = lorad.Lorad(ldcfg["uplink"], ldcfg["downlink"], ldcfg["control"])
    cucfg = cfg['cups']
    cu = cups.CUPSClient(rtcfg['id'], cfg['homedir'])
    dbcfg = cfg['database']

    st = stats.Stats()

    loop = asyncio.get_event_loop()
    # May want to catch other signals too
    signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
    for s in signals:
        loop.add_signal_handler(
            s, lambda s=s: asyncio.create_task(shutdown(loop, s)))

    loop.set_exception_handler(handle_exception)

    try:
        versions = loop.run_until_complete(get_versions(cfg, ld))
        log.debug(f"Versions: {versions}")
        if rtcfg['uri']:
            loop.run_until_complete(rt.connect(rtcfg['uri'],
                                               rtcfg['trust'], rtcfg['crt'],
                                               rtcfg['key'], cucfg['capath']))

        loop.create_task(cups_task(cu, rt, cucfg['uri'],
                                   cucfg['trust'], cucfg['crt'],
                                   cucfg['key'], cucfg['capath'],
                                   cucfg['interval'], versions))
        loop.create_task(ld.recv_task())
        loop.create_task(rt.recv_task())
        loop.create_task(uplink_task(rt, ld, versions, dbcfg))
        loop.create_task(downlink_task(rt, ld, ldcfg))

        loop.create_task(st.recv_task(rt))

        loop.run_forever()
    finally:
        loop.close()
        rt.setLed(False, False)
        log.info("Successfully shutdown station.")


if __name__ == "__main__":
    main()
